from ..layers import (
    SymNetsCategoryLossListInput,
    SymNetsDomainLoss,
    SymNetsEntropyLossListInput,
)
from ..utils import common_functions as c_f
from .base import BaseHook, BaseWrapperHook
from .classification import CLossHook
from .features import FeaturesAndLogitsHook, FeaturesHook
from .losses import TargetEntropyHook
from .mcd import MultipleCLossHook
from .optimizer import OptimizerHook, SummaryHook
from .utils import ChainHook


class SymNetsDomainLossHook(BaseWrapperHook):
    # domain == src or target
    def __init__(self, domain, half_idx, detach_features=False, **kwargs):
        super().__init__(**kwargs)
        self.domain = domain
        self.half_idx = half_idx
        self.hook = FeaturesAndLogitsHook(
            domains=[domain], detach_features=detach_features
        )
        self.loss_fn = SymNetsDomainLoss(half_idx=half_idx)

    def call(self, losses, inputs):
        outputs = self.hook(losses, inputs)[1]
        [logits] = c_f.extract(
            [outputs, inputs], c_f.filter(self.hook.out_keys, f"_logits$")
        )
        loss = self.loss_fn(*logits)
        return {self._loss_keys()[0]: loss}, outputs

    def _loss_keys(self):
        return [f"symnets_{self.domain}_domain_loss_{self.half_idx}"]


class SymNetsCategoryLossHook(CLossHook):
    def __init__(self, **kwargs):
        super().__init__(loss_fn=SymNetsCategoryLossListInput(), **kwargs)

    def _loss_keys(self):
        return ["symnets_category_loss"]


class SymNetsEntropyHook(TargetEntropyHook):
    def __init__(self, **kwargs):
        super().__init__(
            loss_fn=SymNetsEntropyLossListInput(), loss_prefix="symnets_", **kwargs
        )


class SymNetsCHook(BaseWrapperHook):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        c_loss_hook = MultipleCLossHook(detach_features=True)
        d_loss_hook1 = SymNetsDomainLossHook(
            "src", 0, detach_features=True, loss_prefix="c_"
        )
        d_loss_hook2 = SymNetsDomainLossHook(
            "target", 1, detach_features=True, loss_prefix="c_"
        )
        self.hook = ChainHook(c_loss_hook, d_loss_hook1, d_loss_hook2)


class SymNetsGHook(BaseWrapperHook):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        category_loss_hook = SymNetsCategoryLossHook()
        d_loss_hook1 = SymNetsDomainLossHook("target", 0, loss_prefix="g_")
        d_loss_hook2 = SymNetsDomainLossHook("target", 1, loss_prefix="g_")
        entropy_hook = SymNetsEntropyHook()
        self.hook = ChainHook(
            category_loss_hook, d_loss_hook1, d_loss_hook2, entropy_hook
        )


class SymNetsHook(BaseWrapperHook):
    """
    Implementation of
    [Domain-Symmetric Networks for Adversarial Domain Adaptation](https://arxiv.org/abs/1904.04663).
    """

    def __init__(
        self,
        c_opts,
        g_opts,
        c_weighter=None,
        c_reducer=None,
        g_weighter=None,
        g_reducer=None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        f_hook = FeaturesHook()
        c_hook = OptimizerHook(SymNetsCHook(), c_opts, c_weighter, c_reducer)
        g_hook = OptimizerHook(SymNetsGHook(), g_opts, g_weighter, g_reducer)
        s_hook = SummaryHook({"c_loss": c_hook, "g_loss": g_hook})
        self.hook = ChainHook(f_hook, c_hook, g_hook, s_hook)



from .adabn import AdaBNHook
from .adda import ADDAHook
from .aligners import (
    AlignerHook,
    AlignerPlusCHook,
    FeaturesLogitsAlignerHook,
    JointAlignerHook,
    ManyAlignerHook,
)
from .atdoc import ATDOCHook
from .cdan import CDANDomainHookD, CDANDomainHookG, CDANEHook, CDANHook, GANEHook
from .classification import ClassifierHook, CLossHook, FinetunerHook, SoftmaxHook
from .conditions import StrongDHook
from .dann import (
    CDANNEHook,
    DANNEHook,
    DANNHook,
    DANNLogitsHook,
    DANNSoftmaxLogitsHook,
    GradientReversalHook,
    GradientReversalLocallyHook,
    SoftmaxGradientReversalHook,
    SoftmaxGradientReversalLocallyHook,
)
from .domain import DomainLossHook, FeaturesForDomainLossHook
from .domain_confusion import DomainConfusionHook
from .features import (
    CombinedFeaturesHook,
    DLogitsHook,
    FeaturesAndLogitsHook,
    FeaturesHook,
    FeaturesWithGradAndDetachedHook,
    FrozenModelHook,
    LogitsHook,
)
from .gan import GANHook
from .gvb import GVBEHook, GVBGANHook, GVBHook
from .losses import AFNHook, BaseLossHook, BNMHook, BSPHook, MCCHook, TargetEntropyHook
from .mcd import MCDHook, MCDLossHook, MultipleCLossHook
from .reducers import EntropyReducer, MeanReducer, MultipleReducers
from .rtn import ResidualHook, RTNAlignerHook, RTNHook, RTNLogitsHook
from .symnets import (
    SymNetsCategoryLossHook,
    SymNetsCHook,
    SymNetsDomainLossHook,
    SymNetsEntropyHook,
    SymNetsGHook,
    SymNetsHook,
)
from .utils import (
    ApplyToListHook,
    AssertHook,
    ChainHook,
    EmptyHook,
    FalseHook,
    ParallelHook,
    RepeatHook,
    TrueHook,
)
from .vada import VADAHook, VATHook, VATPlusEntropyHook
from .validate import validate_hook



import torch

from ..layers import DiversityLoss
from .base_validator import BaseValidator


class DiversityValidator(BaseValidator):
    """
    Returns the negative of the
    [diversity][pytorch_adapt.layers.diversity_loss.DiversityLoss]
    of all logits.
    """

    def compute_score(self, target_train):
        return -DiversityLoss()(target_train["logits"]).item()



from typing import Any, Dict

import torch
from torch.utils.data import Dataset

from .domain_dataset import DomainDataset


class TargetDataset(DomainDataset):
    """
    This wrapper returns a dictionary,
    but it expects the wrapped dataset to return a tuple of ```(data, label)```.
    """

    def __init__(self, dataset: Dataset, domain: int = 1):
        """
        Arguments:
            dataset: The dataset to wrap
            domain: An integer representing the domain.
        """
        super().__init__(dataset, domain)

    def __getitem__(self, idx: int) -> Dict[str, Any]:
        """
        Returns:
            A dictionary with keys:

            - "target_imgs" (the data)

            - "target_domain" (the integer representing the domain)

            - "target_sample_idx" (idx)
        """

        img, _ = self.dataset[idx]
        return {
            "target_imgs": img,
            "target_domain": self.domain,
            "target_sample_idx": idx,
        }



from typing import Callable

import torch
from pytorch_metric_learning.utils import common_functions as pml_cf
from torchmetrics.functional import accuracy

from ..utils import common_functions as c_f


class SufficientAccuracy(torch.nn.Module):
    """
    Determines if a batch of logits has accuracy greater
    than some threshold. This can be used to control
    program flow.

    Example:
    ```python
    condition_fn = SufficientAccuracy(threshold=0.7)
    if condition_fn(logits, labels):
        ...
    ```
    """

    def __init__(
        self,
        threshold: float,
        accuracy_func: Callable[[torch.Tensor, torch.Tensor], torch.Tensor] = None,
        to_probs_func: Callable[[torch.Tensor], torch.Tensor] = None,
    ):
        """
        Arguments:
            threshold: The accuracy must be greater than this
                for the forward pass to return True.
            accuracy_func: function that takes in ```(to_probs_func(logits), labels)```
                and returns accuracy. If ```None```, then classification accuracy is used.
            to_probs_func: function that processes the logits before they get passed
                to ```accuracy_func```. If ```None```, then ```torch.nn.Sigmoid``` is used
        """

        super().__init__()
        self.threshold = threshold
        self.accuracy_func = c_f.default(accuracy_func, accuracy)
        self.to_probs_func = c_f.default(to_probs_func, torch.nn.Sigmoid())
        pml_cf.add_to_recordable_attributes(
            self, list_of_names=["accuracy", "threshold"]
        )

    def forward(self, x: torch.Tensor, labels: torch.Tensor) -> bool:
        """
        Arguments:
            x: logits to compute accuracy for
            labels: the corresponding labels
        Returns:
            ```True``` if the accuracy is greater than ```self.threshold```
        """
        with torch.no_grad():
            x = self.to_probs_func(x)
            labels = labels.type(torch.int)
            self.accuracy = self.accuracy_func(x, labels).item()
        return self.accuracy > self.threshold

    def extra_repr(self):
        """"""
        return c_f.extra_repr(self, ["threshold"])



from .base_dataset import BaseDataset
from .combined_source_and_target import CombinedSourceAndTargetDataset
from .concat_dataset import ConcatDataset
from .dataloader_creator import DataloaderCreator
from .domainnet import DomainNet, DomainNet126, DomainNet126Full
from .getters import get_mnist_mnistm
from .mnistm import MNISTM
from .office31 import Office31, Office31Full
from .officehome import OfficeHome, OfficeHomeFull
from .pseudo_labeled_dataset import PseudoLabeledDataset
from .source_dataset import SourceDataset
from .target_dataset import TargetDataset



import torch

from ..utils import common_functions as c_f


class SlicedWasserstein(torch.nn.Module):
    """
    Implementation of the loss used in
    [Sliced Wasserstein Discrepancy for Unsupervised Domain Adaptation](https://arxiv.org/abs/1903.04064)
    """

    def __init__(self, m: int = 128):
        """
        Arguments:
            m: The dimensionality to project to.
        """
        super().__init__()
        self.m = 128

    def forward(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        """
        Arguments:
            x: a batch of class predictions
            y: the other batch of class predictions
        Returns:
            The discrepancy between the two batches of class predictions.
        """
        d = x.shape[1]
        proj = torch.randn(d, self.m, device=x.device)
        proj = torch.nn.functional.normalize(proj, dim=0)
        x = torch.matmul(x, proj)
        y = torch.matmul(y, proj)
        x, _ = torch.sort(x, dim=0)
        y, _ = torch.sort(y, dim=0)
        return torch.mean((x - y) ** 2)

    def extra_repr(self):
        """"""
        return c_f.extra_repr(self, ["m"])



import errno
import glob
import itertools
import json
import logging
import os
import re
import shutil
import tarfile
import zipfile
from collections import OrderedDict

import numpy as np
import torch
import tqdm
from pytorch_metric_learning.utils import common_functions as pml_cf

LOGGER_NAME = "PDA"
LOGGER = logging.getLogger(LOGGER_NAME)


def set_logger_name(name):
    global LOGGER_NAME
    global LOGGER
    LOGGER_NAME = name
    LOGGER = logging.getLogger(LOGGER_NAME)


def has_no_parameters(model):
    return len(list(model.parameters())) == 0


def is_optimizer(optimizer):
    return isinstance(optimizer, torch.optim.Optimizer)


def is_none(x):
    return x is None


def default(x, default_x, kwargs=None, condition=None):
    if condition is None:
        condition = is_none
    if kwargs is not None:
        if isinstance(kwargs, dict):
            return default_x(**kwargs) if condition(x) else x
        if isinstance(kwargs, list):
            return default_x(*kwargs) if condition(x) else x
    return default_x if condition(x) else x


def many_default(x, default_x, kwargs=None, condition=None):
    return [default(x[i], default_x[i], kwargs, condition) for i in range(len(x))]


def dict_pop_lazy(x, key, *args, **kwargs):
    y = x.pop(key, None)
    return default(y, *args, **kwargs)


def add_if_new(d, key, x, kwargs, model_name, in_keys, other_args=None):
    # if key is list then assume model returns multiple args
    if not is_list_or_tuple(key) or not is_list_or_tuple(x):
        raise TypeError("key and x must both be a list or tuple")
    condition = is_none
    if any(condition(y) for y in x):
        model = kwargs[model_name]
        input_vals = [kwargs[k] for k in in_keys]
        if other_args is not None:
            input_vals += other_args
        new_x = default(None, model, input_vals, is_none)
        if len(x) > 1:
            if not is_list_or_tuple(new_x) or len(new_x) != len(x):
                raise TypeError(
                    "if input x and key are lists, then output of model must be a list of the same length"
                )
            for i in range(len(x)):
                if condition(x[i]):
                    d[key[i]] = new_x[i]
        else:
            d[key[0]] = new_x


def class_default(cls, x, default):
    return x(cls) if x else default


def list_diff(x, y):
    return list(set(x) - set(y))


def is_list_or_tuple(x):
    return isinstance(x, (list, tuple))


def batch_to_device(batch, device):
    if is_list_or_tuple(batch):
        return [pml_cf.to_device(x, device=device) for x in batch]
    if isinstance(batch, dict):
        return {k: pml_cf.to_device(x, device=device) for k, x in batch.items()}
    return pml_cf.to_device(batch, device=device)


def first_key(in_dict):
    return list(in_dict.keys())[0]


def first_val(in_dict):
    return in_dict[first_key(in_dict)]


def kronecker_product(x, y):
    batch_size = x.shape[0]
    tensor_prod = torch.bmm(x.unsqueeze(2), y.unsqueeze(1))
    return tensor_prod.view(batch_size, -1)


def makedir_if_not_there(dir_name):
    try:
        os.makedirs(dir_name)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise


def filter_kwargs(kwargs, keep):
    return {x: kwargs[x] for x in keep}


def val_dataloader_checks(dataloader):
    if dataloader.drop_last:
        raise ValueError("drop_last should be False when doing validation")
    if not isinstance(dataloader.sampler, torch.utils.data.SequentialSampler):
        raise ValueError("shuffle should be False when doing validation")


def val_collected_data_checks(collected, dataset):
    for k, v in collected.items():
        if len(v) != len(dataset):
            raise ValueError(
                f"The length of {k} should be equal to the length of the validation dataset"
            )


def add_prefix(x, prefix=""):
    prefix = f"{prefix}_" if prefix != "" else prefix
    return f"{prefix}{x}"


def add_suffix(x, suffix=""):
    suffix = f"_{suffix}" if suffix != "" else suffix
    return f"{x}{suffix}"


def class_as_prefix(obj, x, prefix="", suffix=""):
    x = add_prefix(x, cls_name(obj))
    x = add_prefix(x, prefix)
    x = add_suffix(x, suffix)
    return x


def full_path(folder, filename):
    return os.path.join(folder, filename)


def has_state_dict(x):
    return hasattr(x, "state_dict")


def copy_file(src, dst):
    shutil.copyfile(src, dst)


def save_torch_module(x, folder, filename):
    if isinstance(x, torch.nn.DataParallel):
        x = x.module
    makedir_if_not_there(folder)
    filename = full_path(folder, filename)
    LOGGER.debug(f"Saving {filename}")
    pml_cf.save_model(x, filename)


def load_torch_module(x, folder, filename):
    if isinstance(x, torch.nn.DataParallel):
        x = x.module
    filename = full_path(folder, filename)
    LOGGER.debug(f"Loading {filename}")
    pml_cf.load_model(x, filename, torch.device("cuda"))


def save_json(x, folder, filename):
    makedir_if_not_there(folder)
    filename = full_path(folder, filename)
    with open(filename, "w") as fp:
        LOGGER.debug(f"Saving {filename}")
        json.dump(x, fp)


def load_json(folder, filename):
    filename = full_path(folder, filename)
    with open(filename, "r") as fp:
        LOGGER.debug(f"Loading {filename}")
        data = json.load(fp)
    return data


def save_npy(x, folder, filename):
    makedir_if_not_there(folder)
    filename = full_path(folder, filename)
    LOGGER.debug(f"Saving {filename}")
    np.save(filename, x, allow_pickle=False)


def load_npy(folder, filename):
    filename = full_path(folder, filename)
    LOGGER.debug(f"Loading {filename}")
    return np.load(filename, allow_pickle=False)


def enumerate_to_dict(x):
    if isinstance(x, list):
        return {str(i): v for i, v in enumerate(x)}
    elif isinstance(x, dict):
        return x
    else:
        raise TypeError("input must be a list or dict")


def delete_all_but(folder, basename, extension, to_keep):
    basenames_to_keep = []
    for prefix, suffix in to_keep:
        new_basename = add_suffix(add_prefix(basename, prefix), suffix)
        basenames_to_keep.append(new_basename)
    files_to_keep = [full_path(folder, f"{x}{extension}") for x in basenames_to_keep]
    all_matching_files = glob.glob(full_path(folder, f"*{basename}*{extension}"))

    for f in all_matching_files:
        if f not in files_to_keep:
            LOGGER.debug(f"Deleting {f}")
            os.remove(f)


def state_dicts_are_equal(x, y, rtol=None):
    if x.keys() != y.keys():
        return False
    for k, v1 in x.items():
        v2 = y[k]
        if type(v1) != type(v2):
            return False
        if torch.is_tensor(v1):
            if rtol is None and not torch.equal(v1, v2):
                return False
            elif rtol and not torch.allclose(v1, v2, rtol=rtol):
                return False
        elif isinstance(v1, dict) and not state_dicts_are_equal(v1, v2):
            return False
    return True


def set_layers_mode(mode, layer_types=None, layer_names=None):
    if layer_types is None and layer_names is None:
        raise ValueError("layer_types and layer_names cannot both be None")
    if mode not in ["train", "eval"]:
        raise ValueError("mode must be either 'train' or 'eval'")

    def helper(m, mode):
        m.eval() if mode == "eval" else m.train()

    def set_to_mode(m):
        if layer_names is not None:
            classname = cls_name(m)
            if any(L in classname for L in layer_names):
                helper(m, mode)
        if layer_types is not None:
            if isinstance(m, layer_types):
                helper(m, mode)

    return set_to_mode


def batchnorm_types():
    return (torch.nn.BatchNorm1d, torch.nn.BatchNorm2d)


def get_lr(optimizer):
    return optimizer.param_groups[0]["lr"]


def cls_name(x):
    return x.__class__.__name__


def assert_keys_are_present_cls(cls, attr_name, x):
    assert_keys_are_present(getattr(cls, attr_name), attr_name, x)


def assert_keys_are_present(attr, attr_name, x):
    if any(k not in x for k in attr.keys()):
        raise KeyError(f"{attr_name} ({attr}) has keys that are not in input ({x}).")


def get_or_pop(x, k, pop):
    return x.pop(k) if pop else x[k]


def extract(x, keys, pop=False):
    if isinstance(x, list):
        output = []
        for k in keys:
            success = False
            for y in x:
                try:
                    output.append(get_or_pop(y, k, pop))
                    success = True
                    break
                except KeyError:
                    continue
            if not success:
                raise KeyError(f"{k} not found")
        return output
    else:
        return [get_or_pop(x, k, pop) for k in keys]


def filter(x, regex_str, return_order=None):
    r = re.compile(regex_str)
    x = [y for y in x if r.search(y)]
    if return_order:
        output = []
        for s in return_order:
            r = re.compile(s)
            output.extend([y for y in x if r.search(y)])
        return output
    return x


def to_dict(x, keys):
    return {k: x[i] for i, k in enumerate(keys)}


def concat_tensor_dict(x, keys, dim):
    return torch.cat(extract(x, keys), dim=dim)


def zero_loss():
    return torch.tensor(0, dtype=torch.float, requires_grad=True)


def zero_back_step(loss, optimizers):
    for x in optimizers:
        x.zero_grad()
    if loss.grad_fn is not None:
        loss.backward()
        for x in optimizers:
            x.step()


def len_one_tensor(x):
    return x.dim() == 0 or len(x) == 1


def reinit_layer(layer):
    if hasattr(layer, "reset_parameters"):
        layer.reset_parameters()


def reinit(model):
    for L in model.children():
        try:
            for layer in L:
                reinit_layer(layer)
        except TypeError:
            reinit_layer(L)
    return model


def map_keys(x, key_map, must_be_subset=True):
    if must_be_subset and not set(key_map.keys()).issubset(set(x.keys())):
        raise KeyError("key_map keys must be a subset of the input")
    output = {}
    for k, v in x.items():
        new_k = key_map.get(k, k)
        if new_k != k or new_k not in output:
            output[new_k] = v
    return output


# https://stackoverflow.com/a/6117124
def map_keys_substrings(x, key_map):
    rep = dict((re.escape(k), v) for k, v in key_map.items())
    pattern = re.compile("|".join(rep.keys()))
    return pattern.sub(lambda m: rep[re.escape(m.group(0))], x)


def join_lists(x):
    return list(itertools.chain(*x))


def attrs_of_type(cls, obj):
    attrs = [a for a in dir(cls) if not a.startswith("__")]
    attrs = {h: getattr(cls, h) for h in attrs}
    return {k: v for k, v in attrs.items() if isinstance(v, obj)}


def append_error_message(e, msg):
    if len(e.args) >= 1:
        e.args = (e.args[0] + msg,) + e.args[1:]


def requires_grad(x, does=True):
    if isinstance(x, (list, tuple, set)):
        return all(requires_grad(y, does) for y in x)
    if isinstance(x, (dict)):
        return all(requires_grad(y, does) for y in x.values())
    x = x.requires_grad
    return x if does else not x


def to_set(x):
    if isinstance(x, list):
        if all(isinstance(y, dict) for y in x):
            return set().union(*x)
        return set(x)
    return set(x)


def check_domain(cls, domain, keep_len=False):
    if domain is not None:
        if len(torch.unique(domain)) > 1:
            raise ValueError(
                f"{c_f.cls_name(cls)} inference only supports one domain per batch"
            )
        if not keep_len and not len_one_tensor(domain):
            domain = domain[0]
    return domain


def extra_repr(cls, attr_names, delimiter="\n"):
    return delimiter.join([f"{a}={str(getattr(cls, a))}" for a in attr_names])


# copied from https://github.com/pytorch/pytorch/blob/master/torch/nn/modules/module.py
def add_indent(s_, num_spaces=2, not_first=False):
    s = s_.split("\n")
    # don't do anything for single-line stuff
    if len(s) == 1:
        return s_
    if not_first:
        first = s.pop(0)
    s = [(num_spaces * " ") + line for line in s]
    s = "\n".join(s)
    if not_first:
        s = first + "\n" + s
    return s


# copied from https://github.com/pytorch/pytorch/blob/master/torch/nn/modules/module.py
def nice_repr(cls, extra_repr, children):
    # We treat the extra repr like the sub-module, one item per line
    extra_lines = []
    # empty string will be split into list ['']
    if extra_repr:
        extra_lines = extra_repr.split("\n")
    child_lines = []
    for key, value in children.items():
        mod_str = repr(value)
        mod_str = add_indent(mod_str, 2, not_first=True)
        child_lines.append("(" + key + "): " + mod_str)
    lines = extra_lines + child_lines

    main_str = cls_name(cls) + "("
    if lines:
        # simple one-liner info, which most builtin Modules will use
        if len(extra_lines) == 1 and not child_lines:
            main_str += extra_lines[0]
        else:
            main_str += "\n  " + "\n  ".join(lines) + "\n"

    main_str += ")"
    return main_str


def dicts_are_overlapping(x, y, return_overlap=False):
    overlap = x.keys() & y.keys()
    is_overlap = len(overlap) > 0
    if return_overlap:
        return is_overlap, overlap
    return is_overlap


def assert_dicts_are_disjoint(*x):
    for a, b in itertools.combinations(x, 2):
        is_overlap, overlap = dicts_are_overlapping(a, b, True)
        if is_overlap:
            raise KeyError(f"dicts have overlapping keys {overlap}")


def extract_progress(compressed_obj):
    LOGGER.info("Extracting dataset")
    if isinstance(compressed_obj, tarfile.TarFile):
        iterable = compressed_obj
        length = len(compressed_obj.getmembers())
    elif isinstance(compressed_obj, zipfile.ZipFile):
        iterable = compressed_obj.namelist()
        length = len(iterable)
    for member in tqdm.tqdm(iterable, total=length):
        yield member



from typing import Any, Dict, List

import torch
from torch.utils.data import DataLoader

from ..utils import common_functions as c_f


class DataloaderCreator:
    """
    This is a factory class for creating dataloaders.
    The ```__call__``` function takes in keyword arguments which are datasets,
    and outputs a dictionary of dataloaders (one dataloader for each input dataset).
    """

    def __init__(
        self,
        train_kwargs: Dict[str, Any] = None,
        val_kwargs: Dict[str, Any] = None,
        train_names: List[str] = None,
        val_names: List[str] = None,
        all_train: bool = False,
        all_val: bool = False,
        batch_size: int = 32,
        num_workers: int = 0,
    ):
        """
        Arguments:
            train_kwargs: The keyword arguments that will be
                passed to every DataLoader constructor for train-time datasets.
            val_kwargs: The keyword arguments that will be
                passed to every DataLoader constructor for validation-time datasets.
            train_names: A list of the dataset names that are used during training.
            val_names: A list of the dataset names that are used during validation.
            all_train: If True, then all input datasets are assumed to be for training,
                regardless of their names.
            all_val: If True, then all input datasets are assumed to be for validation,
                regardless of their names.
            batch_size: The default ```batch_size``` used in train_kwargs
                (if not provided) and val_kwargs (if not provided)
            num_workers: The default ```num_workers``` used in train_kwargs
                (if not provided) and val_kwargs (if not provided)
        """

        self.train_kwargs = c_f.default(
            train_kwargs,
            {
                "batch_size": batch_size,
                "num_workers": num_workers,
                "shuffle": True,
                "drop_last": True,
            },
        )
        self.val_kwargs = c_f.default(
            val_kwargs,
            {
                "batch_size": batch_size,
                "num_workers": num_workers,
                "shuffle": False,
                "drop_last": False,
            },
        )
        self.train_names = c_f.default(train_names, ["train"])
        self.val_names = c_f.default(
            val_names, ["src_train", "target_train", "src_val", "target_val"]
        )
        if not set(self.train_names).isdisjoint(self.val_names):
            raise ValueError(
                f"train_names {self.train_names} must be disjoint from val_names {self.val_names}"
            )
        if all_train and all_val:
            raise ValueError("all_train and all_val cannot both be True")
        self.all_train = all_train
        self.all_val = all_val

    def __call__(self, **kwargs) -> Dict[str, DataLoader]:
        """
        Arguments:
            **kwargs: keyword arguments mapping from dataset names to datasets.
        Returns:
            a dictionary mapping from dataset names to dataloaders.
        """

        output = {}
        for k, v in kwargs.items():
            if self.all_train:
                dataloader_kwargs = self.train_kwargs
            elif self.all_val:
                dataloader_kwargs = self.val_kwargs
            elif k in self.train_names:
                dataloader_kwargs = self.train_kwargs
            elif k in self.val_names:
                dataloader_kwargs = self.val_kwargs
            else:
                raise ValueError(
                    f"Dataset split name must be in {self.train_names} or {self.val_names}, or one of self.all_train or self.all_val must be true"
                )
            output[k] = torch.utils.data.DataLoader(v, **dataloader_kwargs)
        return output



import torch

from .entropy_loss import entropy


class DiversityLoss(torch.nn.Module):
    """
    Encourages predictions to be uniform, batch wise.
    Takes logits (before softmax) as input.

    For example:

    - A tensor with a large loss: ```torch.tensor([[1e4, 0, 0], [1e4, 0, 0], [1e4, 0, 0]])```

    - A tensor with a small loss: ```torch.tensor([[1e4, 0, 0], [0, 1e4, 0], [0, 0, 1e4]])```
    """

    def forward(self, logits):
        """"""
        logits = torch.mean(logits, dim=0, keepdim=True)
        return -torch.mean(entropy(logits))



import copy

import torch

from ..containers import KeyEnforcer, MultipleContainers, Optimizers
from ..hooks import ADDAHook
from ..utils.common_functions import check_domain
from .base_adapter import BaseAdapter
from .utils import default_optimizer_tuple


class ADDA(BaseAdapter):
    """
    Wraps [ADDAHook][pytorch_adapt.hooks.adda].
    """

    hook_cls = ADDAHook

    def inference_default(self, x, domain):
        domain = check_domain(self, domain)
        fe = "G" if domain == 0 else "T"
        features = self.models[fe](x)
        logits = self.models["C"](features)
        return features, logits

    def get_default_containers(self):
        optimizers = Optimizers(default_optimizer_tuple(), keys=["T", "D"])
        return MultipleContainers(optimizers=optimizers)

    def get_key_enforcer(self):
        return KeyEnforcer(
            models=["G", "C", "D", "T"],
            optimizers=["D", "T"],
        )

    def init_hook(self, hook_kwargs):
        self.hook = self.hook_cls(
            d_opts=[self.optimizers["D"]], g_opts=[self.optimizers["T"]], **hook_kwargs
        )

    def init_containers_and_check_keys(self):
        self.containers["models"]["T"] = copy.deepcopy(self.containers["models"]["G"])
        super().init_containers_and_check_keys()



import torch


def get_silhouette_score(feats, labels):
    device, dtype = feats.device, feats.dtype
    unique_labels = torch.unique(labels)
    num_samples = len(feats)
    if not (1 < len(unique_labels) < num_samples):
        raise ValueError("num unique labels must be > 1 and < num samples")
    scores = []
    for L in unique_labels:
        curr_cluster = feats[labels == L]
        num_elements = len(curr_cluster)
        if num_elements > 1:
            intra_cluster_dists = torch.cdist(curr_cluster, curr_cluster)
            mean_intra_dists = torch.sum(intra_cluster_dists, dim=1) / (
                num_elements - 1
            )  # minus 1 to exclude self distance
            dists_to_other_clusters = []
            for otherL in unique_labels:
                if otherL != L:
                    other_cluster = feats[labels == otherL]
                    inter_cluster_dists = torch.cdist(curr_cluster, other_cluster)
                    mean_inter_dists = torch.sum(inter_cluster_dists, dim=1) / (
                        len(other_cluster)
                    )
                    dists_to_other_clusters.append(mean_inter_dists)
            dists_to_other_clusters = torch.stack(dists_to_other_clusters, dim=1)
            min_dists, _ = torch.min(dists_to_other_clusters, dim=1)
            curr_scores = (min_dists - mean_intra_dists) / (
                torch.maximum(min_dists, mean_intra_dists)
            )
        else:
            curr_scores = torch.tensor([0], device=device, dtype=dtype)

        scores.append(curr_scores)

    scores = torch.cat(scores, dim=0)
    if len(scores) != num_samples:
        raise ValueError(
            f"scores (shape {scores.shape}) should have same length as feats (shape {feats.shape})"
        )
    return torch.mean(scores).item()


class SilhouetteScore(torch.nn.Module):
    """
    A PyTorch implementation of the
    [silhouette score](https://scikit-learn.org/stable/modules/generated/sklearn.metrics.silhouette_score.html)
    """

    def forward(self, features: torch.Tensor, labels: torch.Tensor) -> float:
        """ """
        return get_silhouette_score(features, labels)



import torch

from ..utils import common_functions as c_f


def batch_spectral_loss(x, k):
    singular_values = torch.linalg.svdvals(x)
    return torch.sum(singular_values[:k] ** 2)


class BatchSpectralLoss(torch.nn.Module):
    """
    Implementation of the loss in
    [Transferability vs. Discriminability: Batch Spectral
    Penalization for Adversarial Domain Adaptation](http://proceedings.mlr.press/v97/chen19i.html).
    The loss is the sum of the squares of the first k singular values.
    """

    def __init__(self, k: int = 1):
        """
        Arguments:
            k: the number of singular values to include in the loss
        """
        super().__init__()
        self.k = k

    def forward(self, x):
        """"""
        return batch_spectral_loss(x, self.k)

    def extra_repr(self):
        """"""
        return c_f.extra_repr(self, ["k"])



import torch

from ..utils import common_functions as c_f


class MeanDistLoss(torch.nn.Module):
    def __init__(self, p=2):
        super().__init__()
        self.p = p

    def forward(self, x, y):
        return torch.mean(torch.cdist(x, y, p=self.p))

    def extra_repr(self):
        return c_f.extra_repr(self, ["p"])
